@@ -36,13 +36,14 @@ $actions{permit} = sub ($) {
 $actions{graph} = sub ($) {
   my ($dbh) = @_;
 
-  if (!-e $config{rrdfile}) {
+  if (!-e $config{rrdfile_graylist}) {
     # hmmm, step implies crontab entry here...
-    RRDs::create($config{rrdfile}, "--start", "1000000000", "--step", 300,
-                 "DS:deny:GAUGE:600:U:U", "DS:halfopen:GAUGE:600:0:U",
-                 "DS:permit:GAUGE:600:U:U", "RRA:MAX:0:1:210240");
+    RRDs::create($config{rrdfile_graylist}, "--start", "1000000000",
+                 "--step", 300,
+                 "DS:deny:GAUGE:600:0:U", "DS:halfopen:GAUGE:600:0:U",
+                 "DS:permit:GAUGE:600:0:U", "RRA:MAX:0:1:210240");
     my $err = RRDs::error;
-    die "create $config{rrdfile} failed: $err" if $err;
+    die "create $config{rrdfile_graylist} failed: $err" if $err;
   }
   # get numbers
   my $cnt_deny = dbcount($dbh, "ts_ok IS NULL");
@@ -54,10 +55,10 @@ $actions{graph} = sub ($) {
   if (defined $cnt_deny && defined $cnt_halfopen && defined $cnt_all) {
     $cnt_permit = $cnt_all - $cnt_deny - $cnt_halfopen;
     # update rrd file
-    RRDs::update($config{rrdfile}, "-t", "deny:halfopen:permit",
+    RRDs::update($config{rrdfile_graylist}, "-t", "deny:halfopen:permit",
                  "N:$cnt_deny:$cnt_halfopen:$cnt_permit");
     my $err = RRDs::error;
-    die "update $config{rrdfile} failed: $err" if $err;
+    die "update $config{rrdfile_graylist} failed: $err" if $err;
   } else {
     $cnt_deny = $cnt_halfopen = $cnt_permit = "?";
   }
@@ -66,48 +67,49 @@ $actions{graph} = sub ($) {
     my $fn = $config{picpath} . "/graylist" . $period . ".png";
     RRDs::graph($fn, "-s", "end-" . $period, "-l", "0", "-w", "500",
                 "-t", "Graylisting per " . $period,
-                "DEF:deny=$config{rrdfile}:deny:MAX",
-                "DEF:halfopen=$config{rrdfile}:halfopen:MAX",
-                "DEF:permit=$config{rrdfile}:permit:MAX",
+                "DEF:deny=$config{rrdfile_graylist}:deny:MAX",
+                "DEF:halfopen=$config{rrdfile_graylist}:halfopen:MAX",
+                "DEF:permit=$config{rrdfile_graylist}:permit:MAX",
                 "AREA:permit#00FF00:permit\\: $cnt_permit\\n",
                 "STACK:halfopen#0000FF:half open\\: $cnt_halfopen\\n",
                 "STACK:deny#FF0000:deny\\: $cnt_deny");
     my $err = RRDs::error;
-    die "graph $fn ($config{rrdfile}) failed: $err" if $err;
+    die "graph $fn ($config{rrdfile_graylist}) failed: $err" if $err;
   }
 
-  if (!-e $config{rrdfile2}) {
+  if (!-e $config{rrdfile_mailcount}) {
     # hmmm, step implies crontab entry here...
-    RRDs::create($config{rrdfile2}, "--start", "1000000000", "--step", 300,
-                 "DS:mails:COUNTER:600:U:U", "DS:recipients:COUNTER:600:0:U",
+    RRDs::create($config{rrdfile_mailcount}, "--start", "1000000000",
+                 "--step", 300,
+                 "DS:mails:COUNTER:600:0:U", "DS:recipients:COUNTER:600:0:U",
                  "RRA:MIN:0:1:210240", "RRA:AVERAGE:0:1:210240",
                  "RRA:MAX:0:1:210240", "RRA:LAST:0:1:210240");
     my $err = RRDs::error;
-    die "create $config{rrdfile2} failed: $err" if $err;
+    die "create $config{rrdfile_mailcount} failed: $err" if $err;
   }
   # get numbers
   my $sql = "SELECT COUNT(*), SUM(count) FROM mailcount";
   my ($cnt_mails, $cnt_recipients) = $dbh->selectrow_array($sql);
   if (defined $cnt_mails && defined $cnt_recipients) {
     # update rrd file
-    RRDs::update($config{rrdfile2}, "-t", "mails:recipients",
+    RRDs::update($config{rrdfile_mailcount}, "-t", "mails:recipients",
                  "N:$cnt_mails:$cnt_recipients");
     my $err = RRDs::error;
-    die "update $config{rrdfile2} failed: $err" if $err;
+    die "update $config{rrdfile_mailcount} failed: $err" if $err;
   }
   # make pictures
   foreach my $period("1h", "1d", "1w", "1m", "1y") {
     my $fn = $config{picpath} . "/mailcount" . $period . ".png";
     RRDs::graph($fn, "-s", "end-" . $period, "-l", "0", "-w", "500",
                 "-t", "Mails per " . $period,
-                "DEF:minmails=$config{rrdfile2}:mails:MIN",
-                "DEF:avgmails=$config{rrdfile2}:mails:AVERAGE",
-                "DEF:maxmails=$config{rrdfile2}:mails:MAX",
-                "DEF:lstmails=$config{rrdfile2}:mails:LAST",
-                "DEF:minrcpts=$config{rrdfile2}:recipients:MIN",
-                "DEF:avgrcpts=$config{rrdfile2}:recipients:AVERAGE",
-                "DEF:maxrcpts=$config{rrdfile2}:recipients:MAX",
-                "DEF:lstrcpts=$config{rrdfile2}:recipients:LAST",
+                "DEF:minmails=$config{rrdfile_mailcount}:mails:MIN",
+                "DEF:avgmails=$config{rrdfile_mailcount}:mails:AVERAGE",
+                "DEF:maxmails=$config{rrdfile_mailcount}:mails:MAX",
+                "DEF:lstmails=$config{rrdfile_mailcount}:mails:LAST",
+                "DEF:minrcpts=$config{rrdfile_mailcount}:recipients:MIN",
+                "DEF:avgrcpts=$config{rrdfile_mailcount}:recipients:AVERAGE",
+                "DEF:maxrcpts=$config{rrdfile_mailcount}:recipients:MAX",
+                "DEF:lstrcpts=$config{rrdfile_mailcount}:recipients:LAST",
                 "CDEF:varmails=maxmails,minmails,-",
                 "CDEF:varrcpts=maxrcpts,minrcpts,-",
                 "AREA:minmails",
@@ -125,7 +127,68 @@ $actions{graph} = sub ($) {
                 "GPRINT:maxrcpts:MAX:Max\\:%6.1lf %s",
                 "GPRINT:lstrcpts:LAST:Cur\\:%6.1lf %s\\n");
     my $err = RRDs::error;
-    die "graph $fn ($config{rrdfile2}) failed: $err" if $err;
+    die "graph $fn ($config{rrdfile_mailcount}) failed: $err" if $err;
+  }
+
+  my %codes = (
+    pass      => "00FF00",
+    fail      => "FF0000",
+    softfail  => "0088FF",
+    neutral   => "0044FF",
+    none      => "0000FF",
+    error     => "FF4400",
+    permerror => "FF6600",
+    temperror => "FF9900",
+  );
+
+  foreach my $scope(qw(helo mfrom)) {
+    my $rrdfile = $config{"rrdfile_spf_" . $scope};
+    if (!-e $rrdfile) {
+      # e.g. DS:helo_fail:COUNTER:600:0:U
+      my @ds = map { "DS:" . $_ . ":COUNTER:600:0:U" } keys %codes;
+      # hmmm, step implies crontab entry here...
+      RRDs::create($rrdfile, "--start", "1000000000", "--step", 300, @ds,
+                   "RRA:MIN:0:1:210240", "RRA:AVERAGE:0:1:210240",
+                   "RRA:MAX:0:1:210240", "RRA:LAST:0:1:210240");
+      my $err = RRDs::error;
+      die "create $rrdfile failed: $err" if $err;
+    }
+    # get numbers
+    my (@keys, @vals);
+    my %values = map { $_ => 0 } keys %codes;
+    $sql = "SELECT COUNT(code), code FROM spf WHERE scope=? GROUP BY code";
+    my $sth = $dbh->prepare($sql);
+    $sth->execute($scope);
+    while (my ($cnt, $code) = $sth->fetchrow_array()) {
+      if (defined $codes{$code}) {
+        push @keys, $code;
+        push @vals, $cnt;
+        $values{$code} = $cnt;
+      }
+    }
+    $sth->finish();
+    if (@keys) {
+      # update rrd file
+      RRDs::update($rrdfile, "-t", join(":", @keys), "N:" . join(":", @vals));
+      my $err = RRDs::error;
+      die "update $rrdfile failed: $err" if $err;
+    }
+    # make pictures
+    foreach my $period("1h", "1d", "1w", "1m", "1y") {
+      my (@defs, @graphs);
+      while (my ($code, $color) = each %codes) {
+        push @defs, "DEF:" . $code . "=" . $rrdfile . ":" . $code . ":MAX";
+        push @graphs, (@graphs ? "STACK:" : "AREA:") . $code . "#" . $color .
+                      ":" . $code . "\\: " . $values{$code} . "\\n";
+      }
+        
+      my $fn = $config{picpath} . "/spf_" . $scope . $period . ".png";
+      RRDs::graph($fn, "-s", "end-" . $period, "-l", "0", "-w", "500",
+                  "-t", "SPF " . $scope . " per " . $period,
+                  @defs, @graphs);
+      my $err = RRDs::error;
+      die "graph $fn ($rrdfile) failed: $err" if $err;
+    }
   }
 };
 
