@@ -8,6 +8,7 @@ use Sys::Syslog;
 use POSIX;
 use FindBin qw($Bin);
 use Sys::Hostname;
+use Mail::SPF;
 
 my $HOSTNAME = Sys::Hostname::hostname();
 
@@ -43,10 +44,22 @@ sub normalize_ip ($)
   return $ip;
 }
 
-sub choose_action ($)
+sub choose_action ($@)
 {
-  return "action=dunno\n\n" if $_[0];
-  return "action=defer_if_permit Please try again in several minutes.\n\n";
+  my ($action, $reason) = @_;
+
+  $action = "dunno" if !$action;
+
+  if ($action eq "defer") {
+    return "action=defer_if_permit Please try again in several minutes.\n\n";
+  }
+  if ($action eq "prepend") {
+    return "action=prepend $reason\n\n";
+  }
+  if ($action eq "reject") {
+    return "action=reject" . ($reason ? " $reason" : "") . "\n\n";
+  }
+  return "action=dunno\n\n";
 }
 
 # fatal error
@@ -54,7 +67,7 @@ sub byebye ($@)
 {
   my ($errstr, $dbh) = @_;
 
-  print $socket choose_action(1) if $socket;
+  print $socket choose_action("defer") if $socket;
   if ($dbh) {
     foreach my $sth(grep { defined } @{$dbh->{ChildHandles}}) {
       $sth->finish() if (ref($sth) eq "DBI::st");
@@ -88,7 +101,7 @@ sub listen_socket ($)
   if ($listen =~ /^\//) {
     # unix socket
     unlink($listen);
-    $socket = new IO::Socket::UNIX(
+    $socket = IO::Socket::UNIX->new(
       Local  => $listen,
       Type   => IO::Socket::SOCK_STREAM,
       Listen => $config{backlog}) || byebye("socket(): $listen: $!");
@@ -97,7 +110,7 @@ sub listen_socket ($)
     chmod(oct($config{rights}), $listen) || byebye("chmod(): $listen: $!");
   } else {
     # inet socket
-    $socket = new IO::Socket::INET(
+    $socket = IO::Socket::INET->new(
       LocalAddr  => $listen,
       Proto      => "tcp",
       ReuseAddr  => 1,
@@ -126,14 +139,17 @@ sub _graypol_queries ($) {
 sub _graypol_check ($$$$$$) {
   my ($sth, $params, $ip, $from, $to, $do_insert) = @_;
 
+  my $action = "defer";
+
   # check for entry in graylist
-  my $ok = 0;
   $sth->{select}->execute($ip, $from, $to);
+
   if ($sth->{select}->rows == 1) {
     # found an entry
-    ($ok) = $sth->{select}->fetchrow_array();
+    my ($ts_ok) = $sth->{select}->fetchrow_array();
     # update timestamp
-    $sth->{update}->execute(($ok ? 0 : 1), $ip, $from, $to);
+    $sth->{update}->execute(($ts_ok ? 0 : 1), $ip, $from, $to);
+    $action = "dunno" if $ts_ok;
   } elsif ($do_insert) {
     # did not found an entry -> create it
     local $sth->{insert}->{HandleError} = sub {
@@ -148,7 +164,7 @@ sub _graypol_check ($$$$$$) {
                             $to, 1, $params->{instance});
   }
 
-  return $ok;
+  return $action;
 }
 
 # graylisting callbacks
@@ -159,7 +175,7 @@ sub graypol_queries () {
 sub graypol_check ($$) {
   my ($sth, $params) = @_;
 
-  my $ok = 1;
+  my $action = "dunno";
 
   if (($params->{request} eq "smtpd_access_policy") &&
       ($params->{protocol_state} eq "RCPT")) {
@@ -167,14 +183,14 @@ sub graypol_check ($$) {
     my $from = lc($params->{sender});
     my $to = lc($params->{recipient});
 
-    $ok = _graypol_check($sth, $params,
-                         normalize_ip(lc($params->{client_address})),
-                         $from, $to, 1) ||
-          _graypol_check($sth, $params, "::", "",    $to, 0) ||
-          _graypol_check($sth, $params, "::", $from, $to, 0);
+    $action = _graypol_check($sth, $params,
+                             normalize_ip(lc($params->{client_address})),
+                             $from, $to, 1) ||
+              _graypol_check($sth, $params, "::", "",    $to, 0) ||
+              _graypol_check($sth, $params, "::", $from, $to, 0);
   }
 
-  return $ok;
+  return $action;
 }
 
 # mailcount callbacks
@@ -187,9 +203,14 @@ sub mailcount_queries () {
 }
 
 sub mailcount_check ($$$) {
-  my ($sth, $instances, $params) = @_;
+  my ($sth, $private, $params) = @_;
 
-  return 1 if ($params->{request} ne "smtpd_access_policy");
+  return "dunno" if ($params->{request} ne "smtpd_access_policy");
+
+  if (!$$private) {
+    $$private = {};
+  }
+  my $instances = $$private;
 
   my $from = lc($params->{sender});
   my $to = lc($params->{recipient});
@@ -222,7 +243,7 @@ sub mailcount_check ($$$) {
                         $from, $recipients);
   }
 
-  return 1;
+  return "dunno";
 }
 
 # sent callbacks to bypass graylisting if there's been outgoing mail
@@ -241,7 +262,83 @@ sub sent_check ($$) {
     _graypol_check($sth, $params, "::", $to, $from, 1);
   }
 
-  return 1;
+  return "dunno";
+}
+
+# spf callbacks
+sub spf_queries () {
+  return (
+    "insert" => "INSERT INTO spf (hostname, ip, scope, heloname, sender, " .
+                "code, reason, instance, ts_entry) " .
+                "VALUES (?, ?, ?, ?, ?, ?, ?, ?, now())",
+  );
+}
+
+sub spf_check ($$$) {
+  my ($sth, $private, $params) = @_;
+
+  return "dunno" if ($params->{request} ne "smtpd_access_policy");
+
+  if (!$$private) {
+    $$private = Mail::SPF::Server->new();
+  }
+  my $spf_srv = $$private;
+
+  my $from = lc($params->{sender});
+
+  my $req = Mail::SPF::Request->new(
+    scope         => "mfrom",
+    identity      => $params->{sender},
+    ip_address    => $params->{client_address},
+    helo_identity => $params->{helo_name}
+  );
+  my $res = $spf_srv->process($req);
+
+  $sth->{insert}->execute($HOSTNAME, $params->{client_address}, "mfrom",
+                          $params->{helo_name}, $from, $res->code, $res->text,
+                          $params->{instance});
+  Sys::Syslog::syslog("info", "instance=%s client=%s scope=mfrom " .
+                      "helo=%s from=%s code=%s reason=%s",
+                      $params->{instance}, $params->{client_address},
+                      $params->{helo_name}, $from, $res->code, $res->text);
+
+  if (($res->code eq "fail") || ($res->code eq "permerror")) {
+    return ("reject", $res->local_explanation);
+  } else {
+    return ("prepend", $res->received_spf_header);
+  }
+}
+
+sub spf_helo_check ($$$) {
+  my ($sth, $private, $params) = @_;
+
+  return "dunno" if ($params->{request} ne "smtpd_access_policy");
+
+  if (!$$private) {
+    $$private = Mail::SPF::Server->new();
+  }
+  my $spf_srv = $$private;
+
+  my $req = Mail::SPF::Request->new(
+    scope      => "helo",
+    identity   => $params->{helo_name},
+    ip_address => $params->{client_address}
+  );
+  my $res = $spf_srv->process($req);
+
+  $sth->{insert}->execute($HOSTNAME, $params->{client_address}, "helo",
+                          $params->{helo_name}, undef, $res->code, $res->text,
+                          $params->{instance});
+  Sys::Syslog::syslog("info", "instance=%s client=%s scope=helo " .
+                      "helo=%s code=%s reason=%s",
+                      $params->{instance}, $params->{client_address},
+                      $params->{helo_name}, $res->code, $res->text);
+
+  if (($res->code eq "fail") || ($res->code eq "permerror")) {
+    return ("reject", $res->local_explanation);
+  } else {
+    return ("prepend", $res->received_spf_header);
+  }
 }
 
 
@@ -279,7 +376,7 @@ Sys::Syslog::syslog("warning", "%s", "$err") if $err;
 
 ### create listen sockets and get their file descriptors
 my %listen_sockets;
-foreach (qw(graypol mailcount sent)) {
+foreach (qw(graypol mailcount sent spf spf_helo)) {
   $listen_sockets{$_}->{socket} = listen_socket($config{"listen_" . $_});
   $listen_sockets{$_}->{fileno} = fileno($listen_sockets{$_}->{socket});
 }
@@ -288,7 +385,7 @@ foreach (qw(graypol mailcount sent)) {
 my ($unpriv_uid, $unpriv_gid) = get_uid_gid($config{user});
 
 ### save pid
-if (open(FH, $config{pidfile})) {
+if (open(FH, "<", $config{pidfile})) {
   $pid = <FH>;
   close(FH);
   if (kill(0, $pid)) {
@@ -337,8 +434,9 @@ for ($running = 1; $running; ) {
     next if $pid; # parent continues with next listening socket
 
     # child
+    $0 .= " [" . $name . "]";
     $SIG{TERM} = "DEFAULT";
-    $SIG{INT} = "DEFAULT";
+    $SIG{INT}  = "DEFAULT";
     $SIG{CHLD} = "DEFAULT";
     foreach (keys %listen_sockets) {
       close($listen_sockets{$_}->{socket});
@@ -356,6 +454,8 @@ for ($running = 1; $running; ) {
     %queries = graypol_queries() if ($name eq "graypol");
     %queries = mailcount_queries() if ($name eq "mailcount");
     %queries = sent_queries() if ($name eq "sent");
+    %queries = spf_queries() if ($name eq "spf");
+    %queries = spf_queries() if ($name eq "spf_helo");
 
     my %sth;
     while (my ($stmt_name, $query) = each %queries) {
@@ -363,7 +463,7 @@ for ($running = 1; $running; ) {
     }
 
     # client loop
-    my (%params, %instances);
+    my (%params, $private);
     while (<$socket>) {
       s/\r|\n//sgio;
       if (!/^$/o) {
@@ -373,17 +473,24 @@ for ($running = 1; $running; ) {
         next;
       }
       # empty line, end of parameters
-      my $ok = 1;
+      my $action = "dunno";
+      my $reason;
       if ($name eq "graypol") {
-        $ok = graypol_check(\%sth, \%params);
+        ($action, $reason) = graypol_check(\%sth, \%params);
       }
       if ($name eq "mailcount") {
-        $ok = mailcount_check(\%sth, \%instances, \%params);
+        ($action, $reason) = mailcount_check(\%sth, \$private, \%params);
       }
       if ($name eq "sent") {
-        $ok = sent_check(\%sth, \%params);
+        ($action, $reason) = sent_check(\%sth, \%params);
       }
-      print $socket choose_action($ok);
+      if ($name eq "spf") {
+        ($action, $reason) = spf_check(\%sth, \$private, \%params);
+      }
+      if ($name eq "spf_helo") {
+        ($action, $reason) = spf_helo_check(\%sth, \$private, \%params);
+      }
+      print $socket choose_action($action, $reason);
     }
 
     # child exits
